#ifndef __COMMS_H__
#define __COMMS_H__

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>

/*probably would be better to retrieve LPT port addy...*/
unsigned int LPT1 = 0x378;

int InitGSComms(void);
int CloseGSComms(void);
int GSFcn1(int);
int SendNibble(int);
int ReadWriteNibble(int);
int ReadWriteByte(int);
unsigned int ReadWrite16(unsigned int);
unsigned long ReadWrite32(unsigned long);
int CheckGSPresence(void);
/*common errors*/
int erratta(int code);
int SendCommand(char command);

/*generic commands*/
#define AR_DISCONNECT	'd'
#define AR_END		'e'
#define AR_RAM_DUMP		127


/*****************************************
Generic comms functions courtesy of hcs
http://sourceforge.net/projects/N64dev
nifty, eh?
*****************************************/

/*call to initialize comms
This locks the game and allows PC access
You must re-initialize between read/write to reset command type*/
int InitGSComms(void) {
	int timeout=0x60;/*change this to fiddle with general timeout value*/
	
	while (GSFcn1(3) && timeout) {
		timeout--;
		/*15 second warning...*/
		if(timeout==0x40) printf("\ntimeout in:   %2d",timeout);
		if(timeout<=0x40) printf("\b\b%2d",timeout);
	}
	if (!timeout) erratta(1);
	if (!CheckGSPresence()) erratta(2);
    return 0;
}

/*closing comms cuts the stream.
You only really need to do it when all access is complete
Therefore, read&write all day long, then kill stream at the very end
Note this returns 1 if Shell, 2 if Trainer, 0 if error.*/
int CloseGSComms(void) {
	if (GSFcn1(3)|(!CheckGSPresence())) return 0;
	ReadWriteByte(AR_END);
	return ReadWriteByte(0);
}

int SendCommand(char command) {
	if (InitGSComms()) return 1;
	ReadWriteByte(command);
	return 0;
}

#define ForceDisconnect()	SendCommand(AR_DISCONNECT)

/*checks for sync*/
int GSFcn1(int x) {
	int timeout=0x28;
	unsigned char result=0;

	while (timeout) {
		result<<=4;
		result|=SendNibble(x);
		/* when we recieve a 6, then a 7, we're sync'd*/
		if (result==0x67) break;

		SendNibble(x);
		SendNibble(x>>4);

		timeout--;
	}

	if (!timeout) return 1;
	return 0;
}

int SendNibble(int x) {
	int timeout=0x3e8,retval;

	outport(LPT1,(x & 0xf)|0x10);

	while (timeout && (inport(LPT1+1) & 8)) timeout--;

	retval=((inport(LPT1+1)& 0xf0)^0x80)>>4;

	outport(LPT1,0);

	timeout=0x64;
	while (timeout && (inport(LPT1+1)& 8)) timeout--;

	if (!timeout) return 0;
	else return retval;
}

int ReadWriteNibble(int x) {
	int retval;
	outport(LPT1,(x & 0xf)|0x10);

	while ((~inport(LPT1+1))&8);
	while ((~inport(LPT1+1))&8);

	retval=((inport(LPT1+1)& 0xf0)^0x80)>>4;

	outport(LPT1,0);

	while (inport(LPT1+1)& 8);
	while (inport(LPT1+1)& 8);

	return retval;
}

int ReadWriteByte(int x) {
	return (ReadWriteNibble(x>>4)<<4)|ReadWriteNibble(x);
}

unsigned int ReadWrite16(unsigned int x) {
	return ((ReadWriteNibble(x>>12)<<12)|(ReadWriteNibble(x>>8)<<8)|(ReadWriteNibble(x>>4)<<4)|ReadWriteNibble(x));
}

unsigned long ReadWrite32(unsigned long x) {
	return ((unsigned long)ReadWriteNibble(x>>28)<<28)|\
	((unsigned long)ReadWriteNibble(x>>24)<<24)|\
	((unsigned long)ReadWriteNibble(x>>20)<<20)|\
	((unsigned long)ReadWriteNibble(x>>16)<<16)|\
	((unsigned long)ReadWriteNibble(x>>12)<<12)|\
	((unsigned long)ReadWriteNibble(x>>8)<<8)|\
	((unsigned long)ReadWriteNibble(x>>4)<<4)|\
	(unsigned long)ReadWriteNibble(x);
}


/* this breaks when the GS can allow ram access GS -> gs*/
int CheckGSPresence(void) {
	int timeout=0x3e8,result;

	while (timeout) {
		result=(ReadWriteNibble(0x47>>4)<<4)|ReadWriteNibble(0x47);
		timeout--;
		if (result != 0x67) continue;
		result=(ReadWriteNibble(0x54>>4)<<4)|ReadWriteNibble(0x54);
		if (result == 0x74) break;
	}

/*	printf("\nGS utilities can't be used while in the GS menu.\n");*/
	if (timeout) return 1;
	return 0;
}

/*Generic Ram Download, effective in both shell and trainer.  Size and address should be word-aligned.  0 if okay, 1 if checksum error.
  Automatically closes port when finished*/
int RamDump(FILE * infile, unsigned long size, unsigned long address) {
	unsigned long c=0,x=0,y=0;
	
	address+=3;
	address&=-4;
	size+=3;
	size&=-4;
	
	if (InitGSComms()) return 1;
	ReadWriteByte(AR_RAM_DUMP);
	ReadWrite32(address);
	ReadWrite32(size);
	printf("Writing to file... 0%%");

	for (c=0; c < size; c+=4) {
		y=ReadWrite32(0);
		fwrite(&y,4,1,infile);
		x+=y;
		if (c & 0x400) printf("\b\b\b%2d%%",c*100/size);
		}
	
	/*confirm data valid*/
	x&=0xFF;
	y=ReadWriteByte(0);
	if(x!=y) return 1;
	
	return 0;
}


int erratta(int code)
{if(code<1) return 0;
else if(code==1) fprintf(stderr,"\nTimeout!  Can not interface with I/O port.\n");
else if(code==2) fprintf(stderr,"\nGS menu detected.  Start normal play or exit the GS menus, then try again\n");
else if(code==3) fprintf(stderr,"\nUnable to close comms.\n");
else fprintf(stderr,"\nAborting...\n");
system("PAUSE");
system("CLS");
exit(code);
}
#endif

